#include "nvfs.h"

int nvfs_create(struct inode *dir, struct dentry *dentry, umode_t mode, bool excl)
{
	struct inode *inode;
	int r;

	r = dquot_initialize(dir);
	if (unlikely(r))
		goto ret1;

	inode = nvfs_new_inode(dir, mode, 0, &dentry->d_name);

	if (IS_ERR(inode)) {
		r = PTR_ERR(inode);
		goto ret1;
	}

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r))
		goto ret2;

	d_instantiate_new(dentry, inode);
	return 0;

ret2:
	nvfs_dec_link_count(inode);
	discard_new_inode(inode);
ret1:
	return r;
}

int nvfs_mkdir(struct inode *dir, struct dentry *dentry, umode_t mode)
{
	struct inode *inode;
	int r;

	r = dquot_initialize(dir);
	if (unlikely(r))
		goto ret1;

	nvfs_inc_link_count(dir);

	inode = nvfs_new_inode(dir, S_IFDIR | mode, 0, &dentry->d_name);

	if (IS_ERR(inode)) {
		r = PTR_ERR(inode);
		goto ret2;
	}

	r = nvfs_init_empty_dir(inode);
	if (unlikely(r))
		goto ret3;

	nvfs_inc_link_count(inode);

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r))
		goto ret4;

	d_instantiate_new(dentry, inode);
	return 0;

ret4:
	nvfs_dec_link_count(inode);
ret3:
	nvfs_dec_link_count(inode);
	discard_new_inode(inode);
ret2:
	nvfs_dec_link_count(dir);
ret1:
	return r;
}

static void nvfs_symlink_create_page(struct nvfs_superblock *nvs, void *page, void *aux)
{
	size_t sl = strlen(aux);
	memcpy_flushcache(page, aux, sl + 1);
}

int nvfs_symlink(struct inode *dir, struct dentry *dentry, const char *symname)
{
	struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
	size_t sl = strlen(symname);
	struct inode *inode;
	void *m;
	int r;

	if (sl >= 1UL << nvs->log2_page_size) {
		r = -ENAMETOOLONG;
		goto ret1;
	}

	r = dquot_initialize(dir);
	if (unlikely(r))
		goto ret1;

	inode = nvfs_new_inode(dir, S_IFLNK | S_IRWXUGO, 0, &dentry->d_name);
	if (IS_ERR(inode)) {
		r = PTR_ERR(inode);
		goto ret1;
	}

	if (likely(sl < NVFS_FAST_SYMLINK_LEN)) {
		struct nvfs_inode *nvi = i_to_nmi(inode)->nvi;
		inode->i_link = memcpy(nvi->i_fast_symlink, symname, sl + 1);
		i_to_nmi(inode)->i_flag |= NVFS_FLAG_FAST_SYMLINK;
		nv_store(nvi->i_flag, cpu_to_le32(i_to_nmi(inode)->i_flag));
	}  else {
		m = nvfs_bmap(i_to_nmi(inode), 0, nvfs_symlink_create_page, (void *)symname, NULL);
		if (unlikely(IS_ERR(m))) {
			r = PTR_ERR(m);
			goto ret2;
		}
		inode->i_link = m;
	}
	nvfs_set_inode_size(i_to_nmi(inode), sl);

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r))
		goto ret2;

	d_instantiate_new(dentry, inode);
	return 0;

ret2:
	nvfs_dec_link_count(inode);
	discard_new_inode(inode);
ret1:
	return r;
}

int nvfs_mknod(struct inode *dir, struct dentry *dentry, umode_t mode, dev_t rdev)
{
	struct inode *inode;
	int r;

	r = dquot_initialize(dir);
	if (unlikely(r))
		goto ret1;

	inode = nvfs_new_inode(dir, mode, rdev, &dentry->d_name);
	if (IS_ERR(inode)) {
		r = PTR_ERR(inode);
		goto ret1;
	}

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r))
		goto ret2;

	d_instantiate_new(dentry, inode);
	return 0;

ret2:
	nvfs_dec_link_count(inode);
	discard_new_inode(inode);
ret1:
	return r;
}

int nvfs_unlink(struct inode *dir, struct dentry *dentry)
{
	int r;
	struct inode *inode;
	struct nvfs_memory_inode *nmi;

	r = dquot_initialize(dir);
	if (unlikely(r))
		return r;

	r = nvfs_delete_de(i_to_nmi(dir), &dentry->d_name);
	if (unlikely(r)) {
		if (r == -ENOENT) {
			struct nvfs_superblock *nvs = dir->i_sb->s_fs_info;
			nvfs_error(nvs, "unlink: directory entry for a file '%s' disappeared", dentry->d_name.name);
			r = -EUCLEAN;
		}
		return r;
	}

	wmb();

	inode = d_inode(dentry);
	nmi = i_to_nmi(inode);

	inode->i_ctime = dir->i_ctime;
	nmi->nvi->i_ctime = cpu_to_le64(inode->i_ctime.tv_sec);
	nmi->nvi->i_ctime_nsec = cpu_to_le32(inode->i_ctime.tv_nsec);
	nvfs_dec_link_count(inode);
	return 0;
}

int nvfs_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = d_inode(dentry);
	int r;

	r = nvfs_is_dir_empty(i_to_nmi(inode));
	if (unlikely(r < 0))
		return r;
	if (unlikely(!r))
		return -ENOTEMPTY;

	r = nvfs_unlink(dir, dentry);
	if (unlikely(r < 0))
		return r;

	inode->i_size = 0;
	nvfs_dec_link_count(inode);
	nvfs_dec_link_count(dir);

	return 0;
}

int nvfs_rename(struct inode *old_dir, struct dentry *old_dentry, struct inode *new_dir, struct dentry *new_dentry, unsigned flags)
{
	struct inode *old_inode = d_inode(old_dentry);
	struct inode *new_inode = d_inode(new_dentry);
	struct timespec64 now;
	int r;
	struct nvfs_dir_entry *nve;

	if (unlikely(flags & ~RENAME_NOREPLACE))
		return -EINVAL;

	r = dquot_initialize(old_dir);
	if (unlikely(r))
		return r;

	r = dquot_initialize(new_dir);
	if (unlikely(r))
		return r;

	if (new_inode) {
		if (S_ISDIR(new_inode->i_mode)) {
			r = nvfs_is_dir_empty(i_to_nmi(new_inode));
			if (unlikely(r < 0))
				return r;
			if (unlikely(!r))
				return -ENOTEMPTY;
		}

		now = current_time(new_inode);
		nvfs_update_time(new_inode, &now, S_CTIME);

		nve = nvfs_find_de(i_to_nmi(new_dir), &new_dentry->d_name);
		if (unlikely(IS_ERR(nve)))
			return PTR_ERR(nve);
		if (unlikely(!nve)) {
			struct nvfs_superblock *nvs = old_inode->i_sb->s_fs_info;
			nvfs_error(nvs, "rename: directory entry for a new file '%s' disappeared", new_dentry->d_name.name);
			return -EUCLEAN;
		}
	} else {
		nve = nvfs_prepare_add(new_dentry);
		if (unlikely(IS_ERR(nve)))
			return PTR_ERR(nve);
		if (S_ISDIR(old_inode->i_mode))
			nvfs_inc_link_count(new_dir);
	}

	now = current_time(old_inode);
	nvfs_update_time(old_inode, &now, S_CTIME);

	r = nvfs_delete_de(i_to_nmi(old_dir), &old_dentry->d_name);
	if (unlikely(r)) {
		if (r == -ENOENT) {
			struct nvfs_superblock *nvs = old_inode->i_sb->s_fs_info;
			nvfs_error(nvs, "rename: directory entry for an old file '%s' disappeared", old_dentry->d_name.name);
			r = -EUCLEAN;
		}
		return r;
	}
	wmb();
	nvfs_finish_add(nve, new_dentry, old_inode);

	if (S_ISDIR(old_inode->i_mode))
		nvfs_dec_link_count(old_dir);

	if (new_inode) {
		nvfs_dec_link_count(new_inode);
	}

	return 0;
}

int nvfs_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = d_inode(old_dentry);
	int r;

	r = dquot_initialize(dir);
	if (unlikely(r))
		return r;

	nvfs_inc_link_count(inode);
	ihold(inode);

	r = nvfs_add_link(dentry, inode);
	if (unlikely(r)) {
		nvfs_dec_link_count(inode);
		iput(inode);
		return r;
	}

	d_instantiate(dentry, inode);
	return 0;
}
